// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       binary_generator.cc
 *  \brief      メッセージデータのバイナリ生成クラス
 *  \date       Since: June 28, 2021. 2:04:28 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "binary_generator.h"

#include <algorithm>
#include <list>

#include <json/json.h>

#include <mtk/byte_stream.h>
#include <mtk/csv_parser.h>
#include <mtk/file_pointer.h>
#include <mtk/text_format.h>
#include <mtk/text_stream.h>

#include "message.h"
#include "msg_node/msg_node_parser.h"
#include "xterm_color.h"

namespace msgconv
{
namespace
{
constexpr const char *kAllLanguageName = "all";
constexpr const char *kDefaultFaceName = "default";
constexpr const char *kRubyFaceName = "ruby";
}
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 *  \date       Since: June 28, 2021. 2:05:41 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */
BinaryGenerator::BinaryGenerator() noexcept
    : _hashSeed(MTK::Hash::kFNV1aDefaultValue32)
    , _usingLanguageMap()
    , _languages()
    , _identifiers()
    , _messageTable()
    , _messagePool()
    , _offsetPool()
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      生成の実行
 *  \param[in]  prefix  出力ファイル名のプリフィックス
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::Generate(const RunOptions::Options &options) noexcept
{
    // ハッシュ計算のシード値を設定
    _hashSeed = options.hashSeed;
    PutMessage(MsgID::UseHashSeedValue, _hashSeed);

    // CSVの読み込み
    if (!LoadCSV(options.csvPath, true))
    {
        PutError(MsgID::FailedToReadCSV, options.csvPath);
        return false;
    }

    // インデックス化してメッセージプールを構築
    if (options.indexingType != IndexingType::None)
    {
        // 使用文字を収集
        if (!CollectUsingCharacters(options.indexingType))
        {
            return false;
        }

        // インデックス化したメッセージプールの構築
        if (!BuildIndexedMessagePool(options.indexingType))
        {
            PutError(MsgID::FailedToBuildMessagePool);
            return false;
        }
    }
    // インデックス化せずにメッセージプールを構築
    else
    {
        // メッセージプールの構築
        if (!BuildMessagePool())
        {
            PutError(MsgID::FailedToBuildMessagePool);
            return false;
        }
    }

    // オフセットプールの構築
    if (!BuildOffsetPool())
    {
        PutError(MsgID::FailedToBuildOffsetPool);
        return false;
    }

    // 使用文字リストを出力
    if (options.indexingType != IndexingType::None)
    {
        if (!SaveUsingFile(options))
        {
            return false;
        }
    }

    // バイナリファイルの出力
    auto binaryFilePath = options.outputPath / options.prefix;
    binaryFilePath += ".bin";
    PutMessage(MsgID::OutputBinaryFile, binaryFilePath);
    if (!SaveBinaryFile(binaryFilePath))
    {
        PutError(MsgID::FailedToOutputBinaryFile, binaryFilePath);
        return false;
    }

    // *.incを出力
    if (options.replaceTextPath.empty())
    {
        // メッセージIDを出力
        auto outputPath = options.outputPath / options.prefix;
        outputPath += "_id_hash.inc";
        PutMessage(MsgID::OutputMessageIDList, outputPath);
        if (!SaveMessageID(outputPath))
        {
            PutError(MsgID::FailedToOutputMessageID, outputPath);
            return false;
        }

        // 言語名を出力
        outputPath = options.outputPath / options.prefix;
        outputPath += "_lang_hash.inc";
        PutMessage(MsgID::OutputLanguageNameList, outputPath);
        if (!SaveLanguageID(outputPath))
        {
            PutError(MsgID::FailedToOutputLanguageID, outputPath);
            return false;
        }
    }
    // テキスト置換
    else
    {
        for (const auto &src : options.replaceTextPath)
        {
            TextReplacer replacer(*this);

            replacer.SetCondition(TextReplacer::Condition::Binary, true);
            replacer.SetCondition(TextReplacer::Condition::Indexed, _messagePool.isIndexed);

            if (!replacer.Replace(options.outputPath, src))
            {
                return false;
            }
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      読み込んだCSVの情報を表示
 */
/* ------------------------------------------------------------------------- */
void BinaryGenerator::PrintCSVInfo() const noexcept
{
    // 言語
    printf("Languages:\n");
    for (const auto &lang : _languages)
    {
        printf("  %s: %08X\n", lang.name.c_str(), lang.hash);
    }

    // ID
    printf("Identifier:\n");
    for (const auto &identifier : _identifiers)
    {
        printf("  %s: %08X\n", identifier.name.c_str(), identifier.hash);
    }

    // メッセージ
    for (size_t languageIndex = 0; languageIndex < _languages.size(); languageIndex++)
    {
        const auto &language = _languages[languageIndex];
        printf("%s:\n", language.name.c_str());

        for (size_t idIndex = 0; idIndex < _identifiers.size(); idIndex++)
        { const auto &identifier = _identifiers[idIndex];
            const auto &message = _messageTable[languageIndex][idIndex];
            printf("  %s: \"%s\"\n", identifier.name.c_str(), message.text.c_str());

            for (const auto &node : message.nodes)
            {
                printf("   %s\n", node->GetDebugText().c_str());
            }
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      使用する文字を収集
 *  \param[in]  indexingType    インデックス方法
 *  \retval     true            成功
 *  \retval     false           失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::CollectUsingCharacters(IndexingType indexingType) noexcept
{
    // 全ての言語を1つにまとめる
    if (indexingType == IndexingType::All)
    {
        auto &faceMap = _usingLanguageMap[kAllLanguageName];

        for (size_t languageIndex = 0; languageIndex < _languages.size(); languageIndex++)
        {
            for (const auto &messageData : _messageTable[languageIndex])
            {
                AddUsingCharacter(faceMap, messageData);
            }
        }
    }
    // 言語ごとに収集
    else if (indexingType == IndexingType::EveryLanguage)
    {
        for (size_t languageIndex = 0; languageIndex < _languages.size(); languageIndex++)
        {
            const auto &languageInfo = _languages[languageIndex];
            auto &faceMap = _usingLanguageMap[languageInfo.name];

            for (const auto &messageData : _messageTable[languageIndex])
            {
                AddUsingCharacter(faceMap, messageData);
            }
        }
    }
    else
    {
        return false;
    }

    // インデックスを算出
    for (auto &language : _usingLanguageMap)
    {
        for (auto &face : language.second)
        {
            uint32_t i = 0;
            for (auto &character : face.second)
            {
                character.second.index = i;
                i++;
            }
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージデータの使用文字を追加
 *  \param[out] faceMap     追加先のフェイスマップ
 *  \param[in]  messageData 追加するメッセージデータ
 */
/* ------------------------------------------------------------------------- */
void BinaryGenerator::AddUsingCharacter(BinaryGenerator::UsingFaceMap &faceMap, const BinaryGenerator::MessageData &messageData) const noexcept
{
    std::string currentFace = kDefaultFaceName;

    // 文字追加用のラムダ式
    auto addCharacter = [](UsingCharacterMap &map, const char *text) noexcept
    {
        MTK::TextStream stream(reinterpret_cast<const uint8_t *>(text), strlen(text), MTK::TextEncoding::UTF8);
        for (auto character = stream.Read(); !character.IsNull(); character = stream.Read())
        {
            // 不正な文字は飛ばす
            if (!character.IsValid() || character.IsNewLine())
            {
                continue;
            }
            // それ以外
            else
            {
                auto &characterInfo = map[character.GetCode()];
                if (characterInfo.useCount == 0)
                {
                    characterInfo.code = character.GetCode();
                    characterInfo.name = character.ToString();
                }
                characterInfo.useCount++;
            }
        }
    };

    for (const auto &node : messageData.nodes)
    {
        // テキスト
        if (const auto *textNode = NodeCast<TextMsgNode>(node.get()); textNode != nullptr)
        {
            addCharacter(faceMap[currentFace], textNode->GetText().c_str());
        }
        // 書体変更
        else if (const auto *faceNode = NodeCast<FaceMsgNode>(node.get()); faceNode != nullptr)
        {
            currentFace = faceNode->GetFaceName();
        }
        // タグのリセット
        else if (const auto *resetNode = NodeCast<ResetMsgNode>(node.get()); resetNode != nullptr)
        {
            switch (resetNode->GetResetType())
            {
                case TagResetType::Face:
                case TagResetType::All:
                    currentFace = kDefaultFaceName;
                    break;

                default:
                    break;
            }
        }
        // ルビ
        else if (const auto *rubyNode = NodeCast<RubyMsgNode>(node.get()); rubyNode != nullptr)
        {
            addCharacter(faceMap[currentFace], rubyNode->GetBaseText().c_str());
            addCharacter(faceMap[kRubyFaceName], rubyNode->GetRubyText().c_str());
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      CSVファイルを読み込み
 *  \param[in]  csvPath     CSVファイルのパス
 *  \param[in]  enableParse パースの有効化フラグ
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::LoadCSV(const std::filesystem::path &csvPath, bool enableParse) noexcept
{
    // CSVを読み込み
    MTK::CSVParser csvParser;
    if (!csvParser.Load(csvPath))
    {
        return false;
    }

    // 言語情報を取得
    for (int fieldIndex = 1; fieldIndex < csvParser.GetFieldCount(); fieldIndex++)
    {
        auto language = csvParser.GetField(0, fieldIndex);
        if (language.empty())
        {
            continue;
        }
        
        LanguageInfo info;
        info.name = language;
        info.hash = GetHash(info.name.c_str());
        info.fieldIndex = fieldIndex;

        _languages.push_back(info);
    }

    // メッセージIDのテーブルを取得
    for (int recordIndex = 1; recordIndex < csvParser.GetRecordCount(); recordIndex++)
    {
        auto identifier = csvParser.GetField(recordIndex, 0);
        if (identifier.empty() || identifier[0] == '#')
        {
            continue;
        }

        MessageID msgID;
        msgID.name = identifier;
        msgID.hash = GetHash(msgID.name.c_str());
        msgID.recordIndex = recordIndex;

        _identifiers.push_back(msgID);
    }

    // メッセージIDのテーブルをソート
    std::sort(_identifiers.begin(), _identifiers.end(),
            [](const MessageID &a, const MessageID &b)
            {
                return a.hash < b.hash;
            });

    // ハッシュの衝突チェック
    for (size_t i = 0; i < _identifiers.size() - 1; i++)
    {
        if (_identifiers[i].hash == _identifiers[i + 1].hash)
        {
            PutError(MsgID::HashCollisionDetected, _identifiers[i].name, _identifiers[i + 1].name, _identifiers[i].hash);
        }
    }
    
    // メッセージの内容を読み込み
    for (size_t languageIndex = 0; languageIndex < _languages.size(); languageIndex++)
    {
        const auto &language = _languages[languageIndex];
        _messageTable.push_back(std::vector<MessageData>());

        for (size_t idIndex = 0; idIndex < _identifiers.size(); idIndex++)
        {
            const auto &identifier = _identifiers[idIndex];

            auto message = csvParser.GetField(identifier.recordIndex, language.fieldIndex);

            // 言語のインデックスが0でない場合，メッセージが空白であればインデックス0のメッセージを使い回す
            if ((languageIndex != 0) && message.empty())
            {
                message = csvParser.GetField(identifier.recordIndex, _languages[0].fieldIndex);
            }

            MessageData data;
            data.text = message;
            data.offset = 0;

            if (enableParse)
            {
                ParseMesasge(data.nodes, data.text);
            }

            _messageTable[languageIndex].push_back(std::move(data));
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージプールの構築
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::BuildMessagePool() noexcept
{
    std::vector<char> pool;
    pool.reserve(64 * 1024);

    // メッセージデータを走査
    for (size_t languageIndex = 0; languageIndex < _languages.size(); languageIndex++)
    {
        for (size_t idIndex = 0; idIndex < _identifiers.size(); idIndex++)
        {
            auto &messageData = _messageTable[languageIndex][idIndex];

            // 言語のインデックスが0でない場合，メッセージが空白であればインデックス0のメッセージを使い回す
            if ((languageIndex != 0) && (messageData.text.empty()))
            {
                messageData.offset = _messageTable[0][idIndex].offset;
            }
            // そうでなければテキストをプールに挿入
            else
            {
                messageData.offset = pool.size();
                for (const char c : messageData.text)
                {
                    pool.push_back(c);
                }
                pool.push_back('\0');   // イテレータは終端コードを含まないのでNULL文字を入れておく
            }
        }
    }

    // 生成したvector配列からMessagePoolを設定
    _messagePool.isIndexed = false;
    _messagePool.size = pool.size();
    if (_messagePool.size != 0)
    {
        _messagePool.buffer = std::make_unique<std::byte[]>(_messagePool.size);
        memcpy(_messagePool.buffer.get(), pool.data(), _messagePool.size);
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      インデックス文字列のメッセージプールを構築
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::BuildIndexedMessagePool(IndexingType indexingType) noexcept
{
    std::vector<IndexedCharacter> pool;
    pool.reserve(64 * 1024);

    bool hasError = false;

    for (size_t languageIndex = 0; languageIndex < _languages.size(); languageIndex++)
    {
        const auto &languageInfo = _languages[languageIndex];
        const auto *languageName = (indexingType == IndexingType::All) ? kAllLanguageName : languageInfo.name.c_str();
        const auto &faceMap = _usingLanguageMap[languageName];

        for (size_t idIndex = 0; idIndex < _identifiers.size(); idIndex++)
        {
            auto &messageData = _messageTable[languageIndex][idIndex];
            auto &identifier = _identifiers[idIndex];

            // オフセットを設定
            messageData.offset = pool.size() * sizeof(IndexedCharacter);

            std::string currentFaceName = kDefaultFaceName;

            for (const auto &node : messageData.nodes)
            {
                // テキスト
                if (const auto *textNode = NodeCast<TextMsgNode>(node.get()); textNode != nullptr)
                {
                    if (const auto &map = faceMap.find(currentFaceName); map != faceMap.end())
                    {
                        if (!AddIndexedCharacter(pool, map->second, textNode->GetText()))
                        {
                            hasError = true;
                            break;
                        }
                    }
                    else
                    {
                        PutError(MsgID::DetectUnknownTypefaceName, currentFaceName);
                        hasError = true;
                        break;
                    }
                }
                // 置換
                else if (const auto *substNode = NodeCast<SubstMsgNode>(node.get()); substNode != nullptr)
                {
                    auto indexedOptions = ToIndexedCharactersFromFormatOption(substNode->GetOption());
                    for (const auto index : indexedOptions)
                    {
                        pool.push_back(index);
                    }
                    pool.push_back(kIndexedCharacterSubstitute);
                }
                // 色変更
                else if (const auto *colorNode = NodeCast<ColorMsgNode>(node.get()); colorNode != nullptr)
                {
                    if (const auto xcolor = GetXColor(MTK::Hash::FNV1a(colorNode->GetColorName().c_str())); xcolor)
                    {
                        pool.push_back(kTagParameterColor | uint8_t(xcolor.value()));
                    }
                    else
                    {
                        PutWarning(MsgID::DetectUnsupportedColor, colorNode->GetColorName(),
                                languageInfo.name, identifier.name, messageData.text);
                    }
                }
                // 書体変更
                else if (const auto *faceNode = NodeCast<FaceMsgNode>(node.get()); faceNode != nullptr)
                {
                    if (faceMap.find(faceNode->GetFaceName()) != faceMap.end())
                    {
                        auto faceIndex = ToFaceTypeIndex(faceNode->GetFaceName());
                        if (faceIndex == FaceTypeIndex::Default)
                        {
                            PutWarning(MsgID::DetectUnsupportedTypeface, faceNode->GetFaceName(),
                                    languageInfo.name, identifier.name, messageData.text);
                        }
                        else
                        {
                            pool.push_back(kTagParameterFace | uint8_t(faceIndex));
                            currentFaceName = faceNode->GetFaceName();
                        }
                    }
                    else
                    {
                        PutError(MsgID::DetectUnknownTypefaceName, faceNode->GetFaceName());
                        hasError = true;
                        break;
                    }
                }
                // タグのリセット
                else if (const auto *resetNode = NodeCast<ResetMsgNode>(node.get()); resetNode != nullptr)
                {
                    switch (resetNode->GetResetType())
                    {
                        case TagResetType::Color:
                            pool.push_back(kIndexedCharacterColorReset);
                            break;

                        case TagResetType::Face:
                            currentFaceName = kDefaultFaceName;
                            pool.push_back(kIndexedCharacterFaceReset);
                            break;

                        case TagResetType::All:
                            currentFaceName = kDefaultFaceName;
                            pool.push_back(kIndexedCharacterTagReset);
                            break;

                        case TagResetType::Unknown:
                            PutWarning(MsgID::DetectUnsupportedResetTag, std::string() + resetNode->GetLetter(),
                                    languageInfo.name, identifier.name, messageData.text);
                            break;
                    }
                }
                // ルビ
                else if (const auto *rubyNode = NodeCast<RubyMsgNode>(node.get()); rubyNode != nullptr)
                {
                    // ルビ開始文字
                    pool.push_back(rubyNode->IsWide() ? kIndexedCharacterWideRubyStart : kIndexedCharacterRubyStart);

                    // 親文字
                    if (const auto &map = faceMap.find(currentFaceName); map != faceMap.end())
                    {
                        if (!AddIndexedCharacter(pool, map->second, rubyNode->GetBaseText()))
                        {
                            hasError = true;
                            break;
                        }
                    }
                    else
                    {
                        PutError(MsgID::DetectUnknownTypefaceName, currentFaceName);
                        hasError = true;
                        break;
                    }

                    // ルビ区切り文字
                    pool.push_back(kIndexedCharacterRubyDelimiter);

                    // ルビ文字
                    if (const auto &map = faceMap.find(kRubyFaceName); map != faceMap.end())
                    {
                        if (!AddIndexedCharacter(pool, map->second, rubyNode->GetRubyText()))
                        {
                            hasError = true;
                            break;
                        }
                    }
                    else
                    {
                        PutError(MsgID::DetectUnknownTypefaceName, currentFaceName);
                        hasError = true;
                        break;
                    }

                    // ルビ終了文字
                    pool.push_back(kIndexedCharacterRubyEnd);
                }
            }

            // 最後に終端コードを挿入
            pool.push_back(kIndexedCharacterEndOfText);
        }
    }

    // エラーが発生している場合は失敗
    if (hasError)
    {
        return false;
    }

    // 生成したvector配列からMessagePoolを設定
    _messagePool.isIndexed = true;
    _messagePool.size = pool.size() * sizeof(IndexedCharacter);
    if (_messagePool.size != 0)
    {
        _messagePool.buffer = std::make_unique<std::byte[]>(_messagePool.size);
        memcpy(_messagePool.buffer.get(), pool.data(), _messagePool.size);
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージプールにインデックス文字を追加
 *  \param[out] pool    メッセージプール
 *  \param[in]  map     使用文字マップ
 *  \param[in]  string  追加する文字列
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::AddIndexedCharacter(
        std::vector<IndexedCharacter> &pool,
        const UsingCharacterMap &map,
        const std::string &string) const noexcept
{
    // UTF-32に変換してインデックス化
    MTK::TextStream stream(reinterpret_cast<const uint8_t *>(string.c_str()), strlen(string.c_str()), MTK::TextEncoding::UTF8);
    for (auto character = stream.Read(); !character.IsNull(); character = stream.Read())
    {
        // 不正な文字は飛ばす
        if (!character.IsValid())
        {
            continue;
        }
        // 改行
        else if (character.IsNewLine())
        {
            pool.push_back(kIndexedCharacterNewLine);
        }
        // それ以外
        else
        {
            // インデックスに変換して挿入
            const auto &characterInfo = map.find(character.GetCode());
            if (characterInfo == map.end())
            {
                PutError(MsgID::UnregisterdCharacterDetected, character.ToString(), character.GetCode());
                return false;
            }
            else
            {
                pool.push_back(uint16_t(characterInfo->second.index));
            }
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      オフセットプールを構築
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::BuildOffsetPool() noexcept
{
    _offsetPool.elementCount = _languages.size() * _identifiers.size();
    _offsetPool.buffer = std::make_unique<uint32_t[]>(_offsetPool.elementCount);

    auto *pool = _offsetPool.buffer.get();
    for (size_t idIndex = 0; idIndex < _identifiers.size(); idIndex++)
    {
        for (size_t languageIndex = 0; languageIndex < _languages.size(); languageIndex++)
        {
            auto &messageData = _messageTable[languageIndex][idIndex];

            *pool = static_cast<uint32_t>(messageData.offset);
            pool++;
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      使用文字をファイルに出力
 *  \param[in]  options     実行オプション
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::SaveUsingFile(const RunOptions::Options &options) const noexcept
{
    for (const auto &language : _usingLanguageMap)
    {
        Json::Value rootValue;

        for (const auto &face : language.second)
        {
            auto &faceValue = rootValue[face.first];

            int i = 0;
            for (const auto &characterInfo : face.second)
            {
                auto &value = faceValue[i++];
                value["code"] = characterInfo.second.code;
                value["name"] = characterInfo.second.name;
                value["index"] = characterInfo.second.index;
                value["useCount"] = characterInfo.second.useCount;
            }
        }

        // 言語名は小文字に変換
        std::string languageName = language.first;
        for (auto &c : languageName)
        {
            c = char(std::tolower(c));
        }

        // 保存ファイルをオープン
        std::filesystem::path fileName(options.outputPath);
        fileName /= options.prefix;
        fileName += MTK::Format("_using_{}.json", {languageName}); 

        MTK::File::Pointer fp(fileName, "w");
        if (!fp)
        {
            PutError(MsgID::FailedToOutputUsingFile, fileName);
            return false;
        }

        // 書き込み
        Json::StreamWriterBuilder builder;
        builder.settings_["emitUTF8"] = true;
        std::string saveString = Json::writeString(builder, rootValue);
        fwrite(saveString.c_str(), saveString.size(), 1, fp);

        PutMessage(MsgID::OutputUsingFile, fileName);
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      バイナリファイルに保存
 *  \param[in]  binaryFilePath  出力ファイル名
 *  \retval     true            成功
 *  \retval     false           失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::SaveBinaryFile(const std::filesystem::path &binaryFilePath) const noexcept
{
    // 必要なバッファサイズを計算
    size_t bufferSize = 512;     // ヘッダ＋予備
    bufferSize += _languages.size() * sizeof(uint32_t);
    bufferSize += _identifiers.size() * sizeof(uint32_t);
    bufferSize += _offsetPool.elementCount * _languages.size() * sizeof(uint32_t);
    bufferSize += _messagePool.size;

    // バッファとストリームを用意
    auto buffer = std::make_unique<std::byte[]>(bufferSize);
    MTK::ByteStream stream(buffer.get(), bufferSize);

    // フラグを用意
    uint32_t flags = 0;
    if (_messagePool.isIndexed)
    {
        flags |= kBinaryFlagIndexed;
    }

    // ヘッダの書き込み
    stream.Write(_hashSeed);                                    // ハッシュのシード値
    stream.Write(GetHash("MessageData"));                       // バイナリ識別子
    stream.Write(static_cast<uint32_t>(0));                     // リビジョン情報
    stream.Write(static_cast<uint32_t>(_languages.size()));     // 言語数
    stream.Write(static_cast<uint32_t>(_identifiers.size()));   // ID数
    stream.Write(flags);                                        // フラグ

    // 言語情報の書き込み
    for (const auto &language : _languages)
    {
        stream.Write(language.hash);
    }

    // IDのハッシュ値を書き込み
    for (const auto &identifier : _identifiers)
    {
        stream.Write(identifier.hash);
    }

    // オフセットプールを書き込み
    size_t offsetPoolSize = _offsetPool.elementCount * sizeof(uint32_t);
    stream.Write(static_cast<uint32_t>(offsetPoolSize));
    stream.Write(_offsetPool.buffer.get(), offsetPoolSize);

    // メッセージプールを書き込み
    stream.Write(static_cast<uint32_t>(_messagePool.size));
    stream.Write(_messagePool.buffer.get(), _messagePool.size);

    // 終端マーカーを書き込み
    stream.Write(GetHash("EndOfRecord"));

    // ストリームがオーバーフローしていたら失敗
    if (stream.IsOverflowed())
    {
        PutError(MsgID::OutputBufferOverflowDetected);
        return false;
    }

    // バッファの内容をファイルに出力
    MTK::File::Pointer fp(binaryFilePath, "wb");
    if (!fp)
    {
        return false;
    }
    fwrite(buffer.get(), 1, stream.GetOffset(), fp);

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージIDのハッシュ値を保存
 *  \param[in]  outputPath  出力ファイルのパス
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::SaveMessageID(const std::filesystem::path &outputPath) const noexcept
{
    MTK::File::Pointer fp(outputPath.c_str(), "w");
    if (!fp)
    {
        return false;
    }

    for (const auto &identifier : _identifiers)
    {
        fprintf(fp, "%s = 0x%08X,\n", identifier.name.c_str(), identifier.hash);
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      言語名のハッシュ値を保存
 *  \param[in]  outputPath  出力ファイルのパス
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::SaveLanguageID(const std::filesystem::path &outputPath) const noexcept
{
    MTK::File::Pointer fp(outputPath.c_str(), "w");
    if (!fp)
    {
        return false;
    }

    for (const auto &langage : _languages)
    {
        fprintf(fp, "%s = 0x%08X,\n", langage.name.c_str(), langage.hash);
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          置換コマンドの実行
 *  \param[in,out]  fp          出力先のファイルポインタ
 *  \param[in]      node        コマンドノード
 *  \param[in]      srcFilePath ソースファイルのパス
 *  \retval         true        成功（継続）
 *  \retval         false       失敗（中断）
 */
/* ------------------------------------------------------------------------- */
bool BinaryGenerator::DoReplaceCommand(FILE *fp, const TextReplacer::Node &node, const std::filesystem::path &srcFilePath) noexcept
{
    // ID_LIST コマンド
    if (node.command == TextReplacer::Command::ID_LIST)
    {
        for (const auto &identifier : _identifiers)
        {
            fprintf(fp, "%s%s = 0x%08X,\n", node.indent.c_str(), identifier.name.c_str(), identifier.hash);
        }

        return true;
    }
    // LANGUAGE_LIST コマンド
    else if (node.command == TextReplacer::Command::LANGUAGE_LIST)
    {
        for (const auto &language : _languages)
        {
            fprintf(fp, "%s%s = 0x%08X,\n", node.indent.c_str(), language.name.c_str(), language.hash);
        }

        return true;
    }

    // それ以外は無視
    PutWarning(MsgID::NotWorkCommandWithBinary, srcFilePath, node.lineNumber, node.commandName);
    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      書体名をインデックスに変換
 *  \param[in]  faceName    書体名
 *  \return     対応したインデックス．不明な場合はFaceTypeIndex::Default
 */
/* ------------------------------------------------------------------------- */
FaceTypeIndex BinaryGenerator::ToFaceTypeIndex(const std::string &faceName) const noexcept
{
    if (faceName == "bold")
    {
        return FaceTypeIndex::Bold;
    }

    return FaceTypeIndex::Default;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      テキストフォーマットのオプションをインデックス文字列に変換
 *  \param[in]  option      テキストフォーマットのオプション
 *  \return     インデックス化したオプション
 */
/* ------------------------------------------------------------------------- */
std::vector<IndexedCharacter> BinaryGenerator::ToIndexedCharactersFromFormatOption(const MTK::FormatArgument::Option &option) noexcept
{
    std::vector<IndexedCharacter> indexes;
    indexes.reserve(0xF);
    MTK::FormatArgument::Option defaultOption;
    
    // インデックス
    if (option.index != defaultOption.index)
    {
        indexes.push_back(kSubstituteFormatMaskIndex | static_cast<uint8_t>(option.index));
    }

    // 桁幅
    if (option.width != defaultOption.width)
    {
        indexes.push_back(kSubstituteFormatMaskWidth | static_cast<uint8_t>(option.width));
    }

    // 桁埋め文字
    if (option.fillCharacter != defaultOption.fillCharacter)
    {
        indexes.push_back(kSubstituteFormatMaskFillCharacter | static_cast<uint8_t>(option.fillCharacter));
    }

    // 小数点以下の桁幅
    if (option.precision != defaultOption.precision)
    {
        indexes.push_back(kSubstituteFormatMaskPrecision | static_cast<uint8_t>(option.precision));
    }

    // 16進数の桁幅
    if (option.hexWidth != defaultOption.hexWidth)
    {
        indexes.push_back(kSubstituteFormatMaskHexWidth | static_cast<uint8_t>(option.hexWidth));
    }

    // 各種フラグ
    if ((option.showPlusSign != defaultOption.showPlusSign) ||
        (option.showSeparator != defaultOption.showSeparator) ||
        (option.valueDisplayMode != defaultOption.valueDisplayMode) ||
        (option.skipAutoIndex != defaultOption.skipAutoIndex))
    {
        uint8_t flags = 0;
        
        // 符号の表示
        if (option.showPlusSign)
        {
            flags |= 1 << kSubstituteFlagBitShowPlusSign;
        }
        // 区切り文字の表示
        if (option.showSeparator)
        {
            flags |= 1 << kSubstituteFlagBitShowSeparator;
        }
        // 自動インデックスのスキップ
        if (option.skipAutoIndex)
        {
            flags |= 1 << kSubstituteFlagBitSkipAutoIndex;
        }
        
        // 値の表示モード
        flags |= static_cast<uint8_t>(option.valueDisplayMode) << kSubstituteFlagBitValueDisplayModeLo;

        indexes.push_back(kSubstituteFormatMaskFlags | flags);
    }

    return indexes;
}
}   // namespace msgconv
// vim: et ts=4 sw=4 sts=4
