// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       binary_generator.h
 *  \brief      メッセージデータのバイナリ生成クラス
 *  \date       Since: June 28, 2021. 2:04:29 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_BINARY_GENERATOR_H_1624813469
#define INCGUARD_BINARY_GENERATOR_H_1624813469

#include "indexed_character.h"

#include <cstdint>
#include <list>
#include <map>
#include <memory>
#include <string>
#include <unordered_map>
#include <vector>

#include <mtk/text_format_argument.h>
#include <mtk/text_hash.h>

#include "msg_node/msg_node.h"
#include "run_options.h"
#include "text_replacer.h"

namespace msgconv
{
class BinaryGenerator : public TextReplaceCommandProcessor
{
public:
    BinaryGenerator() noexcept;

    bool Generate(const RunOptions::Options &options) noexcept;

    void PrintCSVInfo() const noexcept;

private:
    //! 使用する文字情報
    struct UsingCharacter
    {
        char32_t code;      //!< 文字コード（UTF-32）
        std::string name;   //!< 名前（文字）
        uint32_t index;     //!< インデックス
        uint32_t useCount;  //!< 使用回数

        UsingCharacter() noexcept
            : code(0)
            , name("")
            , index(0)
            , useCount(0)
        {
        }
    };
    using UsingCharacterMap = std::map<char32_t, UsingCharacter>;
    using UsingFaceMap = std::unordered_map<std::string, UsingCharacterMap>;
    using UsingLanguageMap = std::unordered_map<std::string, UsingFaceMap>;

    //! 言語情報
    struct LanguageInfo
    {
        std::string name;       //!< 言語名
        uint32_t hash;          //!< 言語名のハッシュ値
        int fieldIndex;         //!< CSVのフィールドインデックス
    };

    //! メッセージID
    struct MessageID
    {
        std::string name;       //!< ID名
        uint32_t hash;          //!< ID名のハッシュ値
        int recordIndex;        //!< CSVのレコードインデックス
    };

    //! メッセージデータ
    struct MessageData
    {
        std::string text;       //!< テキスト (UTF-8)
        size_t offset;          //!< メッセージプール上のオフセット
        std::vector<std::unique_ptr<MsgNode>> nodes;
    };

    //! メッセージプール
    struct MessagePool
    {
        bool isIndexed;                         //!< インデックス文字列を格納しているかのフラグ
        std::unique_ptr<std::byte[]> buffer;    //!< メッセージプールのバッファ
        size_t size;                            //!< メッセージプールのバッファサイズ
    };

    //! オフセットプール
    struct OffsetPool
    {
        std::unique_ptr<uint32_t[]> buffer;     //!< オフセットプールのバッファ
        size_t elementCount;                    //!< 要素数
    };

    // バイナリデータのフラグ情報
    static constexpr const uint32_t kBinaryFlagIndexed = 1 << 0;   //!< インデックス文字列かのフラグ

    bool LoadCSV(const std::filesystem::path &csvPath, bool enableParse) noexcept;

    bool BuildMessagePool() noexcept;
    bool BuildIndexedMessagePool(IndexingType indexingType) noexcept;
    bool BuildOffsetPool() noexcept;

    bool AddIndexedCharacter(std::vector<IndexedCharacter> &pool, const UsingCharacterMap &map, const std::string &string) const noexcept;

    bool CollectUsingCharacters(IndexingType indexingType) noexcept;
    void AddUsingCharacter(UsingFaceMap &faceMap, const MessageData &messageData) const noexcept;

    bool SaveUsingFile(const RunOptions::Options &options) const noexcept;
    bool SaveBinaryFile(const std::filesystem::path &binaryFilePath) const noexcept;
    bool SaveMessageID(const std::filesystem::path &outputPath) const noexcept;
    bool SaveLanguageID(const std::filesystem::path &outputPath) const noexcept;

    virtual bool DoReplaceCommand(FILE *fp, const TextReplacer::Node &node, const std::filesystem::path &srcFilePath) noexcept override;

    FaceTypeIndex ToFaceTypeIndex(const std::string &faceName) const noexcept;

    std::vector<IndexedCharacter> ToIndexedCharactersFromFormatOption(const MTK::FormatArgument::Option &option) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      テキストからハッシュ値を取得
     *  \param[in]  text    テキスト
     *  \return     ハッシュ値
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetHash(const char *text) const noexcept
    {
        return MTK::Hash::FNV1a(text, _hashSeed);
    }

    uint32_t _hashSeed;

    UsingLanguageMap _usingLanguageMap;

    std::vector<LanguageInfo> _languages;
    std::vector<MessageID> _identifiers;
    std::vector<std::vector<MessageData>> _messageTable;
    MessagePool _messagePool;
    OffsetPool _offsetPool;
};
}   // namespace msgconv

#endif	// INCGUARD_BINARY_GENERATOR_H_1624813469

// vim: et ts=4 sw=4 sts=4
