// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       msg_node_parser.cc
 *  \brief      メッセージノードの解析
 *  \date       Since: May 7, 2023. 21:53:42 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "msg_node_parser.h"

namespace msgconv
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief          テキストをコミット
 *  \param[out]     nodes   ノードリスト
 *  \param[in,out]  text    テキスト
 */
/* ------------------------------------------------------------------------- */
void CommitText(std::vector<std::unique_ptr<MsgNode>> &nodes, std::string &text) noexcept
{
    if (!text.empty())
    {
        auto textNode = std::make_unique<TextMsgNode>(text);
        nodes.push_back(std::move(textNode));
        text.clear();
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タグ名を取得
 *  \param[out] tagName     タグ名の格納先
 *  \param[in]  text        解析元のテキスト
 *  \return     解析後の次の文字の先頭アドレス．失敗時はtextと同じ
 */
/* ------------------------------------------------------------------------- */
const char *GetTagName(std::string &tagName, const char *text) noexcept
{
    for (size_t i = 0; ; i++)
    {
        auto letter = text[i];
        
        if (letter == '>')
        {
            return &text[i + 1];
        }
        
        switch (letter)
        {
            case '\0':
            case 0x0D:
            case 0x0A:
                return text;
            
            default:
                tagName.push_back(letter);
                break;
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          置換オプションの解析
 *  \param[in,out]  text    テキスト
 *  \param[out]     nodes   ノードリスト
 *  \param[in]      current 解析するテキストの先頭
 *  \return         解析後のテキストの先頭．失敗時はcurrentをそのまま返す
 */
/* ------------------------------------------------------------------------- */
const char *ParseSubsitute(
        std::string &text,
        std::vector<std::unique_ptr<MsgNode>> &nodes,
        const char *current) noexcept
{
    //! 先頭が'{'でなければ置換オプションではない
    if (*current != '{')
    {
        return current;
    }

    // '{'が連続している場合はエスケープ
    if (current[1] == '{')
    {
        text.push_back('{');
        return current + 2;
    }

    MTK::FormatArgument::Option option;

    // 解析に成功していればノードリストに追加
    if (auto *next = MTK::FormatArgument::Parse(option, current); next != current)
    {
        CommitText(nodes, text);
        auto substNode = std::make_unique<SubstMsgNode>(option);
        nodes.push_back(std::move(substNode));

        return next;
    }

    return current;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          色変更タグの解析
 *  \param[in,out]  text    テキスト
 *  \param[out]     nodes   ノードリスト
 *  \param[in]      current 解析するテキストの先頭
 *  \return         解析後のテキストの先頭．失敗時はcurrentをそのまま返す
 */
/* ------------------------------------------------------------------------- */
const char *ParseColorTag(
        std::string &text,
        std::vector<std::unique_ptr<MsgNode>> &nodes,
        const char *current) noexcept
{
    //! 先頭が'<'でなければタグではない
    if (*current != '<')
    {
        return current;
    }

    // '<'が連続している場合はエスケープ
    if (current[1] == '<')
    {
        text.push_back('<');
        return current + 2;
    }
    // 2番目の文字が'#'でなければ色変更タグではない
    else if (current[1] != '#')
    {
        return current;
    }

    // タグ名を取得
    std::string tagName;
    if (auto *next = GetTagName(tagName, &current[2]); next != &current[2])
    {
        CommitText(nodes, text);
        auto node = std::make_unique<ColorMsgNode>(tagName);
        nodes.push_back(std::move(node));

        return next;
    }

    return current;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          書体変更タグの解析
 *  \param[in,out]  text    テキスト
 *  \param[out]     nodes   ノードリスト
 *  \param[in]      current 解析するテキストの先頭
 *  \return         解析後のテキストの先頭．失敗時はcurrentをそのまま返す
 */
/* ------------------------------------------------------------------------- */
const char *ParseFaceTag(
        std::string &text,
        std::vector<std::unique_ptr<MsgNode>> &nodes,
        const char *current) noexcept
{
    //! 先頭が'<'でなければタグではない
    if (*current != '<')
    {
        return current;
    }

    // '<'が連続している場合はエスケープ
    if (current[1] == '<')
    {
        text.push_back('<');
        return current + 2;
    }
    // 2番目の文字が'$'でなければ書体変更タグではない
    else if (current[1] != '$')
    {
        return current;
    }

    // タグ名を取得
    std::string tagName;
    if (auto *next = GetTagName(tagName, &current[2]); next != &current[2])
    {
        CommitText(nodes, text);
        auto node = std::make_unique<FaceMsgNode>(tagName);
        nodes.push_back(std::move(node));

        return next;
    }

    return current;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          リセットタグの解析
 *  \param[in,out]  text    テキスト
 *  \param[out]     nodes   ノードリスト
 *  \param[in]      current 解析するテキストの先頭
 *  \return         解析後のテキストの先頭．失敗時はcurrentをそのまま返す
 */
/* ------------------------------------------------------------------------- */
const char *ParseResetTag(
        std::string &text,
        std::vector<std::unique_ptr<MsgNode>> &nodes,
        const char *current) noexcept
{
    //! 先頭が'<'でなければタグではない
    if (*current != '<')
    {
        return current;
    }

    // '<'が連続している場合はエスケープ
    if (current[1] == '<')
    {
        text.push_back('<');
        return current + 2;
    }
    // 2番目の文字が'/'でなければリセットタグではない
    else if (current[1] != '/')
    {
        return current;
    }

    // タグ名を取得
    std::string tagName;
    if (auto *next = GetTagName(tagName, &current[2]); next != &current[2])
    {
        CommitText(nodes, text);

        if (tagName.empty())
        {
            auto node = std::make_unique<ResetMsgNode>(TagResetType::All, 'a');
            nodes.push_back(std::move(node));
        }
        else
        {
            for (const auto letter : tagName)
            {
                auto type = (letter == '#') ? TagResetType::Color :
                            (letter == '$') ? TagResetType::Face  :
                            (letter == 'a') ? TagResetType::All   :
                            TagResetType::Unknown;

                auto node = std::make_unique<ResetMsgNode>(type, letter);
                nodes.push_back(std::move(node));
            }
        }

        return next;
    }

    return current;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          ルビの解析
 *  \param[in,out]  text    テキスト
 *  \param[out]     nodes   ノードリスト
 *  \param[in]      current 解析するテキストの先頭
 *  \return         解析後のテキストの先頭．失敗時はcurrentをそのまま返す
 */
/* ------------------------------------------------------------------------- */
const char *ParseRuby(
        std::string &text,
        std::vector<std::unique_ptr<MsgNode>> &nodes,
        const char *current) noexcept
{
    //! 先頭が'<'でなければタグではない
    if (*current != '<')
    {
        return current;
    }

    // '<'が連続している場合はエスケープ
    if (current[1] == '<')
    {
        text.push_back('<');
        return current + 2;
    }
    // 2番目の文字が'@'でなければルビではない
    else if (current[1] != '@')
    {
        return current;
    }

    std::string baseText;
    std::string rubyText;
    bool isWide = false;

    // タグ名を取得
    std::string tagName;
    if (auto *next = GetTagName(tagName, &current[2]); next != &current[2])
    {
        // 親文字を取得
        size_t delimiter = 0;
        for (size_t i = 0;; i++)
        {
            // 区切り文字':'をチェック
            if (tagName[i] == ':')
            {
                // ':'が連続していなければ区切り文字
                if (tagName[i + 1] != ':')
                {
                    delimiter = i;
                    break;
                }
                // 連続している場合はエスケープシーケンス
                else
                {
                    baseText.push_back(':');
                    i++;
                }
            }
            // 区切り文字'='をチェック
            else if (tagName[i] == '=')
            {
                // '='が連続していなければ区切り文字
                if (tagName[i + 1] != '=')
                {
                    isWide = true;
                    delimiter = i;
                    break;
                }
                // 連続している場合はエスケープシーケンス
                else
                {
                    baseText.push_back('=');
                    i++;
                }
            }
            // 区切り文字が来る前にタグ名が終了していたらルビ扱いにしない
            else if (tagName[i] == '\0')
            {
                return current;
            }
            else
            {
                baseText.push_back(tagName[i]);
            }
        }

        // ルビ文字を取得
        for (size_t i = delimiter + 1; tagName[i] != '\0'; i++)
        {
            rubyText.push_back(tagName[i]);
        }

        // 親文字とルビ文字のどちらかが空であれば失敗
        if (baseText.empty() || rubyText.empty())
        {
            return current;
        }
        
        // ノードを追加
        CommitText(nodes, text);
        auto node = std::make_unique<RubyMsgNode>(baseText, rubyText, isWide);
        nodes.push_back(std::move(node));

        return next;
    }

    return current;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージの解析
 *  \param[out] nodes   解析結果の格納先
 *  \param[in]  message 解析するメッセージ
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool ParseMesasge(std::vector<std::unique_ptr<MsgNode>> &nodes, const std::string &message) noexcept
{
    std::string text;
    auto *current = message.c_str();

    while (*current != 0)
    {
        bool isFound = false;
        for (auto *parser :
        {
            ParseSubsitute,
            ParseColorTag,
            ParseFaceTag,
            ParseResetTag,
            ParseRuby,
        })
        {
            if (auto *next = parser(text, nodes, current); next != current)
            {
                current = next;
                isFound = true;
                break;
            }
        }

        if (!isFound)
        {
            text.push_back(*current);
            current++;
        }
    }

    CommitText(nodes, text);

    return true;
}
}   // namespace msgconv

// vim: et ts=4 sw=4 sts=4
