// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       msg_node_ruby.h
 *  \brief      メッセージノード - ルビ
 *  \date       Since: May 9, 2023. 1:47:11 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_MSG_NODE_RUBY_H_1683564431
#define INCGUARD_MSG_NODE_RUBY_H_1683564431

#include <string>

#include "msg_node.h"

namespace msgconv
{
class RubyMsgNode : public MsgNode
{
public:
    //! ノードタイプの定義
    static constexpr auto kNodeType = MsgNodeType::Ruby;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     *  \param[in]  baseText    親文字
     *  \param[in]  rubyText    ルビ文字
     *  \param[in]  isWide      幅調整を行うかのフラグ
     */
    /* ------------------------------------------------------------------------- */
    RubyMsgNode(const std::string &baseText, const std::string &rubyText, bool isWide) noexcept
        : _baseText(baseText)
        , _rubyText(rubyText)
        , _isWide(isWide)
    {
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ノードタイプの取得
     *  \return     ノードタイプ
     */
    /* ------------------------------------------------------------------------- */
    virtual MsgNodeType GetNodeType() const noexcept override
    {
        return kNodeType;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      親文字の取得
     *  \return     親文字
     */
    /* ------------------------------------------------------------------------- */
    const std::string &GetBaseText() const noexcept
    {
        return _baseText;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ルビ文字の取得
     *  \return     ルビ文字
     */
    /* ------------------------------------------------------------------------- */
    const std::string &GetRubyText() const noexcept
    {
        return _rubyText;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      幅調整を行うかを取得
     *  \retval     true    幅調整を行う
     *  \retval     false   幅調整を行わない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool IsWide() const noexcept
    {
        return _isWide;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      デバッグ用のテキストを取得
     *  \return     デバッグ用のテキスト
     */
    /* ------------------------------------------------------------------------- */
    virtual std::string GetDebugText() const noexcept override
    {
        return MTK::Format("Ruby: \"{}\" {} \"{}\"", {_baseText, _isWide ? "=" : ":", _rubyText});
    }

private:
    std::string _baseText;
    std::string _rubyText;
    bool _isWide;
};

}   // namespace msgconv

#endif	// INCGUARD_MSG_NODE_RUBY_H_1683564431

// vim: et ts=4 sw=4 sts=4
