// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       config_parser.h
 *  \brief      設定データ読み込みクラス
 *  \date       Since: October 6, 2018. 23:09:02 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_CONFIG_PARSER_H_1538834942
#define INCGUARD_CONFIG_PARSER_H_1538834942

#include <filesystem>
#include <unordered_map>

#include <mtk/text_stream.h>

namespace MTK
{
//! 設定データ読み込みクラス
class ConfigParser
{
public:
    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    ConfigParser() noexcept
        : _valueArray()
        , _currentSection("")
    {
    }

    bool Load(const std::filesystem::path &path) noexcept;
    bool ParseArgs(int argc, const char *argv[], int parseStart = 1) noexcept;
    void Clear() noexcept;

    bool Exists(const char *key) const noexcept;
    std::string GetString(const char *key, const char *defaultValue = nullptr) const noexcept;
    int32_t GetInt(const char *key, int32_t defaultValue = 0) const noexcept;
    float GetFloat(const char *key, float defaultValue = 0.0f) const noexcept;
    bool GetBool(const char *key, bool defaultValue = false) const noexcept;

private:
    //! トークンの種類
    enum class TokenType
    {
        String,         //!< 文字列
        IntValue,       //!< 整数値
        FloatValue,     //!< 小数点数値
        SectionName,    //!< セクション名
        Equal,          //!< イコール
    };

    //! トークン
    struct Token
    {
        TokenType type;         //!< トークンの種類
        std::string name;       //!< 設定値の生文字列
        union
        {
            int32_t intValue;   //!< 整数値（TokenType::IntValueのみ有効）
            float floatValue;   //!< 小数点数値（TokenType::FloatValueのみ有効）
        };

        Token()
            : type(TokenType::String)
            , name("")
            , intValue(0)
        {}
    };
    typedef std::vector<Token> TokenArray;

    bool ParseText(TextStream &textStream) noexcept;
    bool ParseLine(TokenArray &tokenArray, const CharacterArray &line) noexcept;
    bool ParseToken(const TokenArray &tokenArray, std::string section) noexcept;
    void AddConfigValue(const std::string section, const std::string key, const Token &value) noexcept;
    bool IsValidKeyOrValue(TokenType tokenType) noexcept;
    bool IsArgsOptionKey(const char *argv) noexcept;

    std::unordered_map<std::string, Token> _valueArray; //!< 設定値を格納する配列（マップ）
    std::string _currentSection;                        //!< 現在のセクション名
};
}   // namespace MTK


#endif  // INCGUARD_CONFIG_PARSER_H_1538834942

// vim: et ts=4 sw=4 sts=4

