// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       run_options.cc
 *  \brief      実行オプションクラス
 *  \date       Since: October 20, 2018. 22:45:11 JST.
 *  \date       Last Change: October 20, 2018. 23:51:29 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "run_options.h"

#include "message.h"

namespace msgconv
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コマンドライン引数をパース
 *  \param[in]  argc    コマンドライン引数の数
 *  \param[in]  argv    コマンドライン引数
 *  \retval     true    成功
 *  \retval     false   失敗（実行不可）
 */
/* ------------------------------------------------------------------------- */
bool RunOptions::ParseCommandlineArgs(int argc, const char *argv[]) noexcept
{
    // 引数が無かったら簡易ヘルプを表示
    if (argc <= 1)
    {
        _shouldPrintHelp = true;
        return true;
    }

    for (int i = 1; i < argc; i++)
    {
        const char *arg = argv[i];
        const char *nextArg = (i < (argc - 1)) ? argv[i + 1] : nullptr;

        // バージョン情報表示
        if ((strcmp(arg, "-v") == 0 ) || (strcmp(arg, "--version") == 0))
        {
            _shouldPrintVersion = true;
            return true;
        }
        // 簡易ヘルプ表示
        else if (strcmp(arg, "--help") == 0)
        {
            _shouldPrintHelp = true;
            return true;
        }
        // 出力ディレクトリ
        else if ((strcmp(arg, "-o") == 0 ) || (strcmp(arg, "--output") == 0))
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, arg);
                return false;
            }
            _options.outputPath = nextArg;
            i++;
        }
        // プリフィックス
        else if ((strcmp(arg, "-p") == 0 ) || (strcmp(arg, "--prerix") == 0))
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, arg);
                return false;
            }
            _options.prefix = nextArg;
            i++;
        }
        // ソースコードで出力
        else if ((strcmp(arg, "-s") == 0 ) || (strcmp(arg, "--source") == 0))
        {
            _options.generateSource = true;
        }
        // 置換元テキストのパス
        else if (strcmp(arg, "--replace") == 0)
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, arg);
                return false;
            }
            _options.replaceTextPath.push_back(nextArg);
            i++;
        }
        // ハッシュ計算のシード値
        else if (strcmp(arg, "--hash-seed") == 0)
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, arg);
                return false;
            }
            if (!GetValueFromString(_options.hashSeed, nextArg))
            {
                PutError(MsgID::InvalidHashSeed, nextArg);
                return false;
            }
            i++;
        }
        // インデックス化指定
        else if (strcmp(arg, "--indexing") == 0)
        {
            _options.indexingType = IndexingType::EveryLanguage;
        }
        else if (strcmp(arg, "--indexing-all") == 0)
        {
            _options.indexingType = IndexingType::All;
        }
        // CSVファイル名
        else
        {
            if (_options.csvPath.empty())
            {
                _options.csvPath = arg;
            }
            else
            {
                PutError(MsgID::MultipleInputFiles);
                return false;
            }
        }
    }

    // CSVファイル名が指定されていなければ実行できない
    if (_options.csvPath.empty())
    {
        PutError(MsgID::InputFileIsNotSpecified);
        return false;
    }

    _canExecute = true;

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列から値を取得
 *  \param[out] dest    出力先
 *  \param[in]  str     変換元の文字列
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool RunOptions::GetValueFromString(uint32_t &dest, const char *str) const noexcept
{
    try
    {
        dest = static_cast<uint32_t>(std::stoul(str, nullptr, 0));
    }
    catch ([[maybe_unused]] const std::invalid_argument &e)
    {
        return false;
    }
    catch ([[maybe_unused]] const std::out_of_range &e)
    {
        return false;
    }

    return true;
}

}   // namespace msgconv

// vim: et ts=4 sw=4 sts=4
