// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       run_options.h
 *  \brief      実行オプションクラス
 *  \date       Since: October 20, 2018. 22:45:12 JST.
 *  \date       Last Change: October 20, 2018. 23:54:05 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_RUN_OPTIONS_H_1540043112
#define INCGUARD_RUN_OPTIONS_H_1540043112

#include <filesystem>
#include <string>
#include <vector>

#include <mtk/text_hash.h>

namespace msgconv
{
enum class IndexingType
{
    None,
    EveryLanguage,
    All
};

//! 実行オプションクラス
class RunOptions
{
public:
    //! 実行オプション値
    struct Options
    {
        std::filesystem::path csvPath;                          //!< CSVファイルのパス
        std::filesystem::path outputPath;                       //!< ファイルの出力パス
        std::filesystem::path prefix;                           //!< ファイルのプリフィックス
        std::vector<std::filesystem::path> replaceTextPath;     //!< 置換テキストのパス
        bool generateSource;                                    //!< ソースコードで出力
        IndexingType indexingType;                              //!< インデックス化の方法
        uint32_t hashSeed;                                      //!< ハッシュ計算のシード値

        //! 初期値
        Options() noexcept
            : csvPath()
            , outputPath("./")
            , prefix("msg")
            , replaceTextPath()
            , generateSource(false)
            , indexingType(IndexingType::None)
            , hashSeed(MTK::Hash::kFNV1aDefaultValue32)
        {}
    };

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    RunOptions() noexcept
        : _canExecute(false)
        , _shouldPrintVersion(false)
        , _shouldPrintHelp(false)
        , _options()
    {
        return;
    }

    bool ParseCommandlineArgs(int argc, const char *argv[]) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      実行可能かどうかを取得
     *  \retval     true    実行可能
     *  \retval     false   実行不可能
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool CanExecute() const noexcept
    {
        return _canExecute;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      バージョン情報の表示かどうかを取得
     *  \retval     true    バージョン情報を表示する
     *  \retval     false   バージョン情報を表示しない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool ShouldPrintVersion() const noexcept
    {
        return _shouldPrintVersion;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      簡易ヘルプの表示かどうかを取得
     *  \retval     true    簡易ヘルプを表示する
     *  \retval     false   簡易ヘルプを表示しない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool ShouldPrintHelp() const noexcept
    {
        return _shouldPrintHelp;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      オプションの取得
     *  \return     オプション情報
     */
    /* ------------------------------------------------------------------------- */
    const Options &GetOptions() noexcept
    {
        return _options;
    }

private:
    bool GetValueFromString(uint32_t &dest, const char *str) const noexcept;


    bool _canExecute;           //!< 実行可能フラグ
    bool _shouldPrintVersion;   //!< バージョン情報表示フラグ
    bool _shouldPrintHelp;      //!< 簡易ヘルプ表示フラグ

    Options _options;           //!< 実行オプション
};

}   // namespace msgconv

#endif  // INCGUARD_RUN_OPTIONS_H_1540043112

// vim: et ts=4 sw=4 sts=4
