// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       source_generator.cc
 *  \brief      ソースコード埋め込み用データの出力クラス
 *  \date       Since: October 20, 2018. 16:47:30 JST.
 *  \date       Last Change: October 20, 2018. 23:33:23 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "source_generator.h"

#include <stdio.h>

#include <mtk/file_pointer.h>

#include "message.h"


namespace msgconv
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 */
/* ------------------------------------------------------------------------- */
SourceGenerator::SourceGenerator() noexcept
    : _csvParser()
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      生成
 *  \param[in]  options     実行オプション
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool SourceGenerator::Generate(const RunOptions::Options &options) noexcept
{
    // CSVを読み込み
    if (!_csvParser.Load(options.csvPath.c_str()))
    {
        PutError(MsgID::FailedToReadCSV, {options.csvPath});
        return false;
    }

    // ファイルに出力
    if (options.replaceTextPath.empty())
    {
        // 出力プリフィックスを生成
        auto outputPrefix = options.outputPath / options.prefix;

        // IDテーブルを生成
        if (!GenerateIDList(_csvParser, outputPrefix))
        {
            PutError(MsgID::FailedToCreateIDList, {outputPrefix});
            return false;
        }

        // メッセージテーブルを生成
        for (int fieldIndex = 1; fieldIndex < _csvParser.GetFieldCount(); fieldIndex++)
        {
            if (!GenerateMessageList(_csvParser, outputPrefix.c_str(), fieldIndex))
            {
                continue;
            }
        }
    }
    // 置換
    else
    {
        for (const auto &srcFilePath : options.replaceTextPath)
        {
            TextReplacer replacer(*this);

            replacer.SetCondition(TextReplacer::Condition::Source, true);

            if (!replacer.Replace(options.outputPath, srcFilePath))
            {
                return false;
            }
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      IDテーブルを生成
 *  \param[in]  csvParser       CSVパーサー
 *  \param[in]  outputPrefix    出力ファイルのプリフィックス（パス含む）
 *  \retval     true            成功
 *  \retval     false           失敗
 */
/* ------------------------------------------------------------------------- */
bool SourceGenerator::GenerateIDList(const MTK::CSVParser &csvParser, const std::filesystem::path &outputPrefix) const noexcept
{
    std::filesystem::path outputPath = outputPrefix;
    outputPath += "_id.inc";

    PutMessage(MsgID::OutputMessageIDList, {outputPath});

    // 出力ファイルをオープン
    MTK::File::Pointer fp(outputPath, "w");
    if (!fp)
    {
        return false;
    }

    return PutIDList(fp, csvParser);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メッセージテーブルを生成
 *  \param[in]  csvParser       CSVパーサー
 *  \param[in]  outputPrefix    出力ファイルのプリフィックス（パス含む）
 *  \param[in]  fieldIndex      フィールドのインデックス
 *  \retval     true            成功
 *  \retval     false           失敗
 */
/* ------------------------------------------------------------------------- */
bool SourceGenerator::GenerateMessageList(const MTK::CSVParser &csvParser, const std::filesystem::path &outputPrefix, int fieldIndex) const noexcept
{
    // 言語タイプを取得
    std::string langType = csvParser.GetField(0, fieldIndex);
    if (langType.empty())
    {
        return false;
    }

    // 小文字に変換
    for (auto &c : langType)
    {
        c = static_cast<char>(std::tolower(c));
    }

    auto outputPath = outputPrefix;
    outputPath += "_";
    outputPath += langType;
    outputPath += ".inc";

    PutMessage(MsgID::OutputMessageTable, {outputPath});

    // 出力ファイルをオープン
    MTK::File::Pointer fp(outputPath, "w");
    if (!fp)
    {
        return false;
    }

    return PutMessageList(fp, csvParser, fieldIndex);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          IDリストを出力
 *  \param[in,out]  fp          出力先のファイルポインタ
 *  \param[in]      csvParser   CSVパーサ
 *  \param[in]      indent      インデント
 *  \retval         true        成功
 *  \retval         false       失敗
 */
/* ------------------------------------------------------------------------- */
bool SourceGenerator::PutIDList(FILE *fp, const MTK::CSVParser &csvParser, const std::string &indent) const noexcept
{
    // フィールドを走査して出力
    for (int recordIndex = 1; recordIndex < csvParser.GetRecordCount(); recordIndex++)
    {
        std::string identifier = csvParser.GetField(recordIndex, 0);
        if (identifier.empty())
        {
            continue;
        }

        fprintf(fp, "%s%s,\n", indent.c_str(), identifier.c_str());
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          メッセージリストを出力
 *  \param[in,out]  fp          出力先のファイルポインタ
 *  \param[in]      csvParser   CSVパーサ
 *  \param[in]      fieldIndex  フィールドのインデックス
 *  \param[in]      indent      インデント
 *  \retval         true        成功
 *  \retval         false       失敗
 */
/* ------------------------------------------------------------------------- */
bool SourceGenerator::PutMessageList(FILE *fp, const MTK::CSVParser &csvParser, int fieldIndex, const std::string &indent) const noexcept
{
    // フィールドを走査して出力
    for (int recordIndex = 1; recordIndex < csvParser.GetRecordCount(); recordIndex++)
    {
        // IDを取得して空であれば飛ばす
        std::string identifier = csvParser.GetField(recordIndex, 0);
        if (identifier.empty())
        {
            continue;
        }

        // そのフィールドが空白であればプライマリのフィールドを使用する
        std::string field = csvParser.GetField(recordIndex, fieldIndex);
        if (field.empty())
        {
            field = csvParser.GetField(recordIndex, 1);
        }

        // IDをコメントとして出力しておく（いらないかも）
        fprintf(fp, "%s/* %s */ ", indent.c_str(), identifier.c_str());

        // フィールドの内容を出力
        fputs("\"", fp);
        for (auto &c : field)
        {
            fprintf(fp, "\\x%02X", static_cast<uint8_t>(c));
        }
        fputs("\",\n", fp);
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          置換コマンドの実行
 *  \param[in,out]  fp          出力先のファイルポインタ
 *  \param[in]      node        コマンドノード
 *  \param[in]      srcFilePath ソースファイルのパス
 *  \retval         true        成功（継続）
 *  \retval         false       失敗（中断）
 */
/* ------------------------------------------------------------------------- */
bool SourceGenerator::DoReplaceCommand(FILE *fp, const TextReplacer::Node &node, const std::filesystem::path &srcFilePath) noexcept
{
    // ID_LIST コマンド
    if (node.command == TextReplacer::Command::ID_LIST)
    {
        return PutIDList(fp, _csvParser, node.indent);
    }
    // MESSAGE_LIST コマンド
    else if (node.command == TextReplacer::Command::MESSAGE_LIST)
    {
        for (int fieldIndex = 1; fieldIndex < _csvParser.GetFieldCount(); fieldIndex++)
        {
            if (_csvParser.GetField(0, fieldIndex) == node.argument)
            {
                return PutMessageList(fp, _csvParser, fieldIndex, node.indent);
            }
        }

        // 言語が見つからなければエラー
        PutError(MsgID::LanguageNotFound, {srcFilePath, node.lineNumber, node.argument});
        return false;
    }
    // LANGUAGE_LIST コマンド
    else if (node.command == TextReplacer::Command::LANGUAGE_LIST)
    {
        for (int fieldIndex = 1; fieldIndex < _csvParser.GetFieldCount(); fieldIndex++)
        {
            fprintf(fp, "%s%s,\n", node.indent.c_str(), _csvParser.GetField(0, fieldIndex).c_str());
        }

        return true;
    }

    // それ以外は無視
    PutWarning(MsgID::NotWorkCommandWithSource, {srcFilePath, node.lineNumber, node.commandName});
    return true;
}
}   // namespace msgconv

// vim: et ts=4 sw=4 sts=4
