// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       text_replacer.cc
 *  \brief      テキスト置換クラス
 *  \date       Since: February 15, 2023. 15:26:18 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <stack>

#include <mtk/file_pointer.h>

#include "text_replacer.h"
#include "message.h"

namespace msgconv
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 */
/* ------------------------------------------------------------------------- */
TextReplacer::TextReplacer(TextReplaceCommandProcessor &commandProcessor) noexcept
    : _commandProcessor(commandProcessor)
    , _nodes()
    , _isBinary(false)
    , _isSource(false)
    , _isIndexed(false)
    , _outputName()
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      条件式で判別する値を設定
 *  \param[in]  condition   条件
 *  \param[in]  value       値
 */
/* ------------------------------------------------------------------------- */
void TextReplacer::SetCondition(Condition condition, bool value) noexcept
{
    switch (condition)
    {
        case Condition::Binary:     _isBinary = value;      break;
        case Condition::Source:     _isSource = value;      break;
        case Condition::Indexed:    _isIndexed = value;     break;

        // これらは無視
        case Condition::AlwaysTrue:
        case Condition::AlwaysFalse:
            break;
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      置換
 *  \param[in]  srcFilePath     変換元のファイルのパス
 *  \retval     true            成功
 *  \retval     false           失敗
 */
/* ------------------------------------------------------------------------- */
bool TextReplacer::Replace(const std::filesystem::path &outputDirectory, const std::filesystem::path &srcFilePath) noexcept
{
    // 入力ファイルをパース
    if (!Parse(srcFilePath))
    {
        return false;
    }

    // 出力ファイル名を決定
    std::filesystem::path destFilePath;
    if (_outputName.empty())
    {
        // 出力ファイル名が指定されていなければ入力ファイル名を使用
        destFilePath = outputDirectory / srcFilePath.filename();
    }
    else
    {
        // 出力ファイル名が絶対パスで指定されていればそのまま使用
        if (_outputName.is_absolute())
        {
            destFilePath = _outputName;
        }
        // 出力ファイル名が相対パスで指定されていれば出力ディレクトリからの相対パス
        else
        {
            destFilePath = outputDirectory / _outputName;
        }
    }
    // 入力ファイルと出力ファイルが等価であれば出力ファイル名に".replaced"を付与する
    if (std::filesystem::exists(destFilePath))
    {
        if (std::filesystem::equivalent(srcFilePath, destFilePath))
        {
            destFilePath += ".replaced";
        }
    }

    // メッセージ表示
    PutMessage(MsgID::DoTextReplace, {srcFilePath, destFilePath});

    // 出力ファイルをオープン
    MTK::File::Pointer destFile(destFilePath, "w");
    if (!destFile)
    {
        PutError(MsgID::FailedToOpenFile, {destFilePath});
        return false;
    }

    // 置換コマンドのノードリストを実行
    for (const auto &node : _nodes)
    {
        switch (node.command)
        {
            // IF
            case Command::IF:
                fprintf(destFile, "%s#if %d   /* auto generated: IF:%s */\n", node.indent.c_str(), IsConsistent(node.condition), node.argument.c_str());
                break;

            // ELIF
            case Command::ELIF:
                fprintf(destFile, "%s#elif %d /* auto generated: ELIF:%s */\n", node.indent.c_str(), IsConsistent(node.condition), node.argument.c_str());
                break;

            // ELSE
            case Command::ELSE:
                fprintf(destFile, "%s#else   /* auto generated */\n", node.indent.c_str());
                break;

            // ENDIF
            case Command::ENDIF:
                fprintf(destFile, "%s#endif  /* auto generated */\n", node.indent.c_str());
                break;

            // ジェネレータクラスに渡すコマンド
            case Command::MESSAGE_LIST:
            case Command::ID_LIST:
            case Command::LANGUAGE_LIST:
                if (!_commandProcessor.DoReplaceCommand(destFile, node, srcFilePath))
                {
                    destFile.Close();
                    std::filesystem::remove(destFilePath);
                    return false;
                }
                break;

            // そのまま出力
            case Command::PUT:
                fputs(node.argument.c_str(), destFile);
                break;

            // 実行しないコマンド（パースの時点で弾いている）
            case Command::NOP:
            case Command::OUTPUT_NAME:
                break;
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      入力ファイルのパース
 *  \param[in]  srcFilePath 入力ファイルのパス
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool TextReplacer::Parse(const std::filesystem::path &srcFilePath) noexcept
{
    // 元になるファイルをオープン
    MTK::File::Pointer srcFile(srcFilePath, "r");
    if (!srcFile)
    {
        PutError(MsgID::FailedToReadFile, {srcFilePath});
        return false;
    }

    // 1行あたりのバッファを準備
    constexpr const size_t kLineBufferSize = 4096;
    char rawLine[kLineBufferSize];

    uint32_t lineNumber = 0;

    while (fgets(rawLine, static_cast<int>(kLineBufferSize), srcFile) != nullptr)
    {
        const char *top = rawLine;

        Node node;
        node.lineNumber = ++lineNumber;

        // インデントを取得
        top = GetIndent(node.indent, top);

        // コマンド名を解釈してノードを追加
        if (GetCommandName(node.commandName, node.argument, top))
        {
            node.command = GetCommand(node.commandName);

            // 不明なコマンド名
            if (node.command == Command::NOP)
            {
                PutError(MsgID::UnknownCommandName, {srcFilePath, lineNumber, node.commandName});
                return false;
            }
            // 出力名
            else if (node.command == Command::OUTPUT_NAME)
            {
                if (node.argument.empty())
                {
                    PutError(MsgID::NotSpecifiedOutputName, {srcFilePath, lineNumber});
                    return false;
                }

                _outputName = node.argument;
            }
            // それ以外はノードリストに追加
            else
            {
                _nodes.push_back(node);
            }
        }
        // コマンド名が指定されていない場合は行全体を引数にしてPUTコマンドを生成
        else
        {
            node.command = Command::PUT;
            node.argument = rawLine;
            node.indent.clear();

            _nodes.push_back(node);
        }
    }

    // ノードリスト構築後の処理
    for (auto &node : _nodes)
    {
        switch (node.command)
        {
            // IF, ELIFコマンドの条件チェック
            case Command::IF:
            case Command::ELIF:
                if (IsValidConditionName(node.argument))
                {
                    node.condition = Condition{MTK::Hash::FNV1a(node.argument.c_str())};
                }
                else
                {
                    PutWarning(MsgID::UnknownConditionName, {srcFilePath, node.lineNumber, node.argument});
                    node.condition = Condition::AlwaysFalse;
                }
                break;

            default:
                break;
        };
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      インデントを取得
 *  \param[out] indent      その行のインデント
 *  \param[in]  text        その行のテキスト
 *  \return     インデントの次の文字
 */
/* ------------------------------------------------------------------------- */
const char *TextReplacer::GetIndent(std::string &indent, const char *text) const noexcept
{
    indent.clear();

    auto *top = text;
    while (*top != '\0')
    {
        // スペースとタブをインデントとして扱う
        if (auto c = *top; (c == ' ') || (c == '\t'))
        {
            indent += c;
        }
        else
        {
            break;
        }

        top++;
    }

    return top;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コマンド名を取得
 *  \param[out] commandName     コマンド名
 *  \param[out] argument        引数名
 *  \param[in]  text            取得元のテキスト
 *  \retval     true            成功
 *  \retval     false           失敗
 */
/* ------------------------------------------------------------------------- */
bool TextReplacer::GetCommandName(std::string &commandName, std::string &argument, const char *text) const noexcept
{
    commandName.clear();
    argument.clear();

    // 先頭が '$' でなければコマンドではない
    if (*text != '$')
    {
        return false;
    }

    bool hasArgs = false;

    // コマンド名の部分を取得
    auto *top = text + 1;
    for (;;)
    {
        if (*top == ':')
        {
            hasArgs = true;
            break;
        }
        else if (*top == '$')
        {
            break;
        }
        else if (*top == '\0')
        {
            return false;
        }

        commandName += *top;

        top++;
    }

    // 引数を取得
    if (hasArgs)
    {
        top++;
        for (;;)
        {
            if (*top == '$')
            {
                break;
            }
            else if (*top == '\0')
            {
                return false;
            }

            argument += *top;

            top++;
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コマンド名からコマンドを取得
 *  \param[in]  commandName     コマンド名
 *  \return     対応するコマンド
 */
/* ------------------------------------------------------------------------- */
TextReplacer::Command TextReplacer::GetCommand(const std::string &commandName) const noexcept
{
    auto command = Command{MTK::Hash::FNV1a(commandName.c_str())};

    switch (command)
    {
        case Command::PUT:
        case Command::OUTPUT_NAME:
        case Command::IF:
        case Command::ELIF:
        case Command::ELSE:
        case Command::ENDIF:
        case Command::MESSAGE_LIST:
        case Command::ID_LIST:
        case Command::LANGUAGE_LIST:
            return command;

        case Command::NOP:
            break;
    }

    return Command::NOP;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      条件名が有効かを取得
 *  \param[in]  conditionName   条件名
 *  \retval     true            有効
 *  \retval     false           無効
 */
/* ------------------------------------------------------------------------- */
bool TextReplacer::IsValidConditionName(const std::string &conditionName) const noexcept
{
    auto condition = Condition{MTK::Hash::FNV1a(conditionName.c_str())};

    switch (condition)
    {
        case Condition::AlwaysFalse:
        case Condition::AlwaysTrue:
        case Condition::Indexed:
        case Condition::Source:
        case Condition::Binary:
            return true;
    }

    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      条件が成立しているかを取得
 *  \param[in]  condition   条件
 *  \retval     true        成立
 *  \retval     false       不成立
 */
/* ------------------------------------------------------------------------- */
bool TextReplacer::IsConsistent(Condition condition) const noexcept
{
    switch (condition)
    {
        case Condition::AlwaysFalse:    return false;
        case Condition::AlwaysTrue:     return true;
        case Condition::Indexed:        return _isIndexed;
        case Condition::Source:         return _isSource;
        case Condition::Binary:         return _isBinary;
    }

    return false;
}
}   // namespace msgconv

// vim: et ts=4 sw=4 sts=4
