// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       text_replacer.h
 *  \brief      テキスト置換クラス
 *  \date       Since: February 15, 2023. 15:26:20 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_TEXT_REPLACER_H_1676442380
#define INCGUARD_TEXT_REPLACER_H_1676442380

#include <filesystem>
#include <list>
#include <string>
#include <vector>

#include <mtk/text_hash.h>

namespace msgconv
{
class TextReplaceCommandProcessor;

//! テキスト置換クラス
class TextReplacer
{
public:
    //! 条件タイプ
    enum class Condition : uint32_t
    {
        AlwaysFalse = MTK::Hash::FNV1a("FALSE"),        //!< 常に不成立
        AlwaysTrue  = MTK::Hash::FNV1a("TRUE"),         //!< 常に成立
        Binary      = MTK::Hash::FNV1a("BINARY"),       //!< バイナリ出力時に成立
        Source      = MTK::Hash::FNV1a("SOURCE"),       //!< ソース出力時に成立
        Indexed     = MTK::Hash::FNV1a("INDEXED"),      //!< インデックス文字列出力時に成立
    };

    //! 置換コマンド
    enum class Command : uint32_t
    {
        NOP             = MTK::Hash::FNV1a("NOP"),              //!< 何もしない
        OUTPUT_NAME     = MTK::Hash::FNV1a("OUTPUT_NAME"),      //!< 出力ファイル名の指定
        PUT             = MTK::Hash::FNV1a("PUT"),              //!< 出力
        IF              = MTK::Hash::FNV1a("IF"),               //!< #if に置き換え
        ELIF            = MTK::Hash::FNV1a("ELIF"),             //!< #elif に置き換え
        ELSE            = MTK::Hash::FNV1a("ELSE"),             //!< #else に置き換え
        ENDIF           = MTK::Hash::FNV1a("ENDIF"),            //!< #endif に置き換え
        MESSAGE_LIST    = MTK::Hash::FNV1a("MESSAGE_LIST"),     //!< メッセージリストに置き換え
        ID_LIST         = MTK::Hash::FNV1a("ID_LIST"),          //!< IDリストに置き換え
        LANGUAGE_LIST   = MTK::Hash::FNV1a("LANGUAGE_LIST"),    //!< 言語リストに置き換え
    };

    //! ノード
    struct Node
    {
        Command command;            //!< コマンド
        Condition condition;        //!< 条件
        std::string commandName;    //!< コマンド名
        std::string argument;       //!< 引数
        std::string indent;         //!< インデント
        uint32_t lineNumber;        //!< 行番号

        //! コンストラクタ
        Node() noexcept
            : command(Command::NOP)
            , condition(Condition::AlwaysFalse)
            , commandName()
            , argument()
            , indent()
            , lineNumber(0)
        {
        }
    };

    TextReplacer(TextReplaceCommandProcessor &commandProcessor) noexcept;

    void SetCondition(Condition condition, bool value) noexcept;
    bool Replace(const std::filesystem::path &outputDirectory, const std::filesystem::path &srcFilePath) noexcept;

private:
    bool Parse(const std::filesystem::path &srcFilePath) noexcept;
    const char *GetIndent(std::string &indent, const char *text) const noexcept;
    bool GetCommandName(std::string &commandName, std::string &argument, const char *text) const noexcept;
    Command GetCommand(const std::string &commandName) const noexcept;
    bool IsValidConditionName(const std::string &conditionName) const noexcept;
    bool IsConsistent(Condition condition) const noexcept;

    TextReplaceCommandProcessor &_commandProcessor;

    std::list<Node> _nodes;

    bool _isBinary;
    bool _isSource;
    bool _isIndexed;

    std::filesystem::path _outputName;
};


//! テキスト置換コマンドの実行インターフェース
class TextReplaceCommandProcessor
{
public:
    virtual ~TextReplaceCommandProcessor() = default;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief          置換コマンドの実行
     *  \param[in,out]  fp          出力先のファイルポインタ
     *  \param[in]      node        コマンドノード
     *  \param[in]      srcFilePath ソースファイルのパス
     *  \retval         true        成功（継続）
     *  \retval         false       失敗（中断）
     */
    /* ------------------------------------------------------------------------- */
    virtual bool DoReplaceCommand(FILE *fp, const TextReplacer::Node &node, const std::filesystem::path &srcFilePath) noexcept = 0;
};

}   // namespace msgconv


#endif	// INCGUARD_TEXT_REPLACER_H_1676442380

// vim: et ts=4 sw=4 sts=4
